@@ -2,8 +2,9 @@
 #define _INSIDE_PYGOBJECT_
 #include "pygobject.h"
 
-static GHashTable *class_hash;
+static PyObject *gerror_exc = NULL;
 
+static GQuark pygobject_class_key = 0;
 static GQuark pygobject_wrapper_key = 0;
 static GQuark pygobject_ownedref_key = 0;
 
@@ -54,7 +55,7 @@ pyg_type_wrapper_repr(PyGTypeWrapper *self)
     char buf[80];
     const gchar *name = g_type_name(self->type);
 
-    g_snprintf(buf, sizeof(buf), "<GType %s>",
+    g_snprintf(buf, sizeof(buf), "<GType %s (%p)>",
 	       name?name:"invalid", self->type);
     return PyString_FromString(buf);
 }
@@ -238,9 +239,6 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
     PyObject *o;
     const char *class_name, *s;
 
-    if (!class_hash)
-	class_hash = g_hash_table_new(g_str_hash, g_str_equal);
-
     class_name = type->tp_name;
     s = strrchr(class_name, '.');
     if (s != NULL)
@@ -267,9 +265,11 @@ pygobject_register_class(PyObject *dict, const gchar *type_name,
 	o = pyg_type_wrapper_new(gtype);
 	PyDict_SetItemString(type->tp_dict, "__gtype__", o);
 	Py_DECREF(o);
-    }
 
-    g_hash_table_insert(class_hash, g_strdup(type_name), type);
+	/* stash a pointer to the python class with the GType */
+	Py_INCREF(type);
+	g_type_set_qdata(gtype, pygobject_class_key, type);
+    }
 
     PyDict_SetItemString(dict, (char *)class_name, (PyObject *)type);
 }
@@ -289,8 +289,8 @@ pygobject_lookup_class(GType gtype)
     PyTypeObject *type;
 
     /* find the python type for this object.  If not found, use parent. */
-    while ((type = g_hash_table_lookup(class_hash, g_type_name(gtype))) == NULL
-           && gtype != 0)
+    while (gtype != G_TYPE_INVALID &&
+	   (type = g_type_get_qdata(gtype, pygobject_class_key)) == NULL)
         gtype = g_type_parent(gtype);
     g_assert(type != NULL);
     return type;
@@ -727,58 +727,92 @@ pyg_value_from_pyobject(GValue *value, PyObject *obj)
 	g_value_set_boolean(value, PyObject_IsTrue(obj));
 	break;
     case G_TYPE_INT:
-	if ((tmp = PyNumber_Int(obj)))
-	    g_value_set_int(value, PyInt_AsLong(tmp));
-	else {
+	g_value_set_int(value, PyInt_AsLong(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
 	    PyErr_Clear();
 	    return -1;
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_UINT:
-	if ((tmp = PyNumber_Int(obj)))
-	    g_value_set_uint(value, PyInt_AsLong(tmp));
-	else {
-	    PyErr_Clear();
-	    return -1;
+	{
+	    if (PyInt_Check(obj)) {
+		glong val;
+
+		val = PyInt_AsLong(obj);
+		if (val >= 0 && val <= G_MAXUINT)
+		    g_value_set_uint(value, (guint)val);
+		else
+		    return -1;
+	    } else {
+		g_value_set_uint(value, PyLong_AsUnsignedLong(obj));
+		if (PyErr_Occurred()) {
+		    g_value_unset(value);
+		    PyErr_Clear();
+		    return -1;
+		}
+	    }
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_LONG:
-	if ((tmp = PyNumber_Int(obj)))
-	    g_value_set_long(value, PyInt_AsLong(tmp));
-	else {
+	g_value_set_long(value, PyInt_AsLong(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
 	    PyErr_Clear();
 	    return -1;
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_ULONG:
-	if ((tmp = PyNumber_Int(obj)))
-	    g_value_set_ulong(value, PyInt_AsLong(tmp));
-	else {
+	{
+	    if (PyInt_Check(obj)) {
+		glong val;
+
+		val = PyInt_AsLong(obj);
+		if (val >= 0)
+		    g_value_set_ulong(value, (gulong)val);
+		else
+		    return -1;
+	    } else {
+		g_value_set_ulong(value, PyLong_AsUnsignedLong(obj));
+		if (PyErr_Occurred()) {
+		    g_value_unset(value);
+		    PyErr_Clear();
+		    return -1;
+		}
+	    }
+	}
+	break;
+    case G_TYPE_INT64:
+	g_value_set_int64(value, PyLong_AsLongLong(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
+	    PyErr_Clear();
+	    return -1;
+	}
+	break;
+    case G_TYPE_UINT64:
+	g_value_set_uint64(value, PyLong_AsUnsignedLongLong(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
 	    PyErr_Clear();
 	    return -1;
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_FLOAT:
-	if ((tmp = PyNumber_Float(obj)))
-	    g_value_set_float(value, PyFloat_AsDouble(tmp));
-	else {
+	g_value_set_float(value, PyFloat_AsDouble(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
 	    PyErr_Clear();
 	    return -1;
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_DOUBLE:
-	if ((tmp = PyNumber_Float(obj)))
-	    g_value_set_double(value, PyFloat_AsDouble(tmp));
-	else {
+	g_value_set_double(value, PyFloat_AsDouble(obj));
+	if (PyErr_Occurred()) {
+	    g_value_unset(value);
 	    PyErr_Clear();
 	    return -1;
 	}
-	Py_DECREF(tmp);
 	break;
     case G_TYPE_STRING:
 	if ((tmp = PyObject_Str(obj)))
@@ -864,11 +898,43 @@ pyg_value_as_pyobject(const GValue *value)
     case G_TYPE_INT:
 	return PyInt_FromLong(g_value_get_int(value));
     case G_TYPE_UINT:
-	return PyInt_FromLong(g_value_get_uint(value));
+	{
+	    gulong val = (gulong) g_value_get_uint(value);
+
+	    if (val <= G_MAXLONG)
+		return PyInt_FromLong((glong) val);
+	    else
+		return PyLong_FromUnsignedLong(val);
+	}
     case G_TYPE_LONG:
 	return PyInt_FromLong(g_value_get_long(value));
     case G_TYPE_ULONG:
-	return PyInt_FromLong(g_value_get_ulong(value));
+	{
+	    gulong val = g_value_get_ulong(value);
+
+	    if (val <= G_MAXLONG)
+		return PyInt_FromLong((glong) val);
+	    else
+		return PyLong_FromUnsignedLong(val);
+	}
+    case G_TYPE_INT64:
+	{
+	    gint64 val = g_value_get_int64(value);
+
+	    if (G_MINLONG <= val && val <= G_MAXLONG)
+		return PyInt_FromLong((glong) val);
+	    else
+		return PyLong_FromLongLong(val);
+	}
+    case G_TYPE_UINT64:
+	{
+	    guint64 val = g_value_get_uint64(value);
+
+	    if (val <= G_MAXLONG)
+		return PyInt_FromLong((glong) val);
+	    else
+		return PyLong_FromUnsignedLongLong(val);
+	}
     case G_TYPE_FLOAT:
 	return PyFloat_FromDouble(g_value_get_float(value));
     case G_TYPE_DOUBLE:
@@ -975,10 +1041,8 @@ pyg_closure_marshal(GClosure *closure,
     }
     ret = PyObject_CallObject(pc->callback, params);
     if (ret == NULL) {
-	if (!pyg_fatal_exceptions_notify()) {
-	    PyErr_Print();
-	    PyErr_Clear();
-	}
+	PyErr_Print();
+	PyErr_Clear();
 	return;
     }
     if (return_value)
@@ -1088,10 +1152,8 @@ pyg_signal_class_closure_marshal(GClosure *closure,
 
     ret = PyObject_CallObject(method, params);
     if (ret == NULL) {
-	if (!pyg_fatal_exceptions_notify()) {
-	    PyErr_Print();
-	    PyErr_Clear();
-	}
+	PyErr_Print();
+	PyErr_Clear();
 	/* XXXX - clean up if threading was used */
 	Py_DECREF(method);
 	return;
@@ -1633,6 +1695,75 @@ pygobject_stop_emission(PyGObject *self, PyObject *args)
     return Py_None;
 }
 
+static PyObject *
+pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
+{
+    GSignalInvocationHint *ihint;
+    guint signal_id, i, len;
+    PyObject *first, *py_ret;
+    gchar *name;
+    GSignalQuery query;
+    GValue *params, ret = { 0, };
+
+    ihint = g_signal_get_invocation_hint(self->obj);
+    signal_id = ihint->signal_id;
+    name = g_signal_name(signal_id);
+
+    len = PyTuple_Size(args);
+    if (signal_id == 0) {
+	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	return NULL;
+    }
+    g_signal_query(signal_id, &query);
+    if (len != query.n_params) {
+	gchar buf[128];
+
+	g_snprintf(buf, sizeof(buf),
+		   "%d parameters needed for signal %s; %d given",
+		   query.n_params, name, len);
+	PyErr_SetString(PyExc_TypeError, buf);
+	return NULL;
+    }
+    params = g_new0(GValue, query.n_params + 1);
+    g_value_init(&params[0], G_OBJECT_TYPE(self->obj));
+    g_value_set_object(&params[0], G_OBJECT(self->obj));
+
+    for (i = 0; i < query.n_params; i++)
+	g_value_init(&params[i + 1],
+		     query.param_types[i] & ~G_SIGNAL_TYPE_STATIC_SCOPE);
+    for (i = 0; i < query.n_params; i++) {
+	PyObject *item = PyTuple_GetItem(args, i);
+
+	if (pyg_value_from_pyobject(&params[i+1], item) < 0) {
+	    gchar buf[128];
+
+	    g_snprintf(buf, sizeof(buf),
+		"could not convert type %s to %s required for parameter %d",
+		item->ob_type->tp_name,
+		g_type_name(G_VALUE_TYPE(&params[i+1])), i);
+	    PyErr_SetString(PyExc_TypeError, buf);
+	    for (i = 0; i < query.n_params + 1; i++)
+		g_value_unset(&params[i]);
+	    g_free(params);
+	    return NULL;
+	}
+    }
+    if (query.return_type != G_TYPE_NONE)
+	g_value_init(&ret, query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE);
+    g_signal_chain_from_overridden(params, &ret);
+    for (i = 0; i < query.n_params + 1; i++)
+	g_value_unset(&params[i]);
+    g_free(params);
+    if (query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE != G_TYPE_NONE) {
+	py_ret = pyg_value_as_pyobject(&ret);
+	g_value_unset(&ret);
+    } else {
+	Py_INCREF(Py_None);
+	py_ret = Py_None;
+    }
+    return py_ret;
+}
+
 static PyMethodDef pygobject_methods[] = {
     { "__gobject_init__", (PyCFunction)pygobject__gobject_init__,
       METH_VARARGS|METH_KEYWORDS },
@@ -1654,6 +1785,7 @@ static PyMethodDef pygobject_methods[] = {
     { "emit", (PyCFunction)pygobject_emit, METH_VARARGS },
     { "stop_emission", (PyCFunction)pygobject_stop_emission, METH_VARARGS },
     { "emit_stop_by_name", (PyCFunction)pygobject_stop_emission,METH_VARARGS },
+    { "chain", (PyCFunction)pygobject_chain_from_overridden,METH_VARARGS },
     { NULL, NULL, 0 }
 };
 
@@ -1925,10 +2057,450 @@ pyg_type_interfaces (PyObject *self, PyObject *args)
     return NULL;
 }
 
+static void
+pyg_object_set_property (GObject *object, guint property_id,
+			 const GValue *value, GParamSpec *pspec)
+{
+    PyObject *object_wrapper, *retval;
+    PyObject *py_pspec, *py_value;
+
+    object_wrapper = pygobject_new(object);
+    g_return_if_fail(object_wrapper != NULL);
+
+    py_pspec = pyg_param_spec_new(pspec);
+    py_value = pyg_value_as_pyobject (value);
+    retval = PyObject_CallMethod(object_wrapper, "do_set_property",
+				 "OO", py_pspec, py_value);
+    if (retval) {
+	Py_DECREF(retval);
+    } else {
+	PyErr_Print();
+	PyErr_Clear();
+    }
+    Py_DECREF(py_pspec);
+    Py_DECREF(py_value);
+}
+
+static void
+pyg_object_get_property (GObject *object, guint property_id,
+			 GValue *value, GParamSpec *pspec)
+{
+    PyObject *object_wrapper, *retval;
+    PyObject *py_pspec;
+
+    object_wrapper = pygobject_new(object);
+    g_return_if_fail(object_wrapper != NULL);
+
+    py_pspec = pyg_param_spec_new(pspec);
+    retval = PyObject_CallMethod(object_wrapper, "do_get_property",
+				 "O", py_pspec);
+    if (retval == NULL || pyg_value_from_pyobject(value, retval) < 0) {
+	PyErr_Print();
+	PyErr_Clear();
+    }
+    Py_XDECREF(retval);
+    Py_DECREF(py_pspec);
+}
+
+static void
+pyg_object_class_init(GObjectClass *class, PyObject *py_class)
+{
+    class->set_property = pyg_object_set_property;
+    class->get_property = pyg_object_get_property;
+}
+
+static gboolean
+create_signal (GType instance_type, const gchar *signal_name, PyObject *tuple)
+{
+    GSignalFlags signal_flags;
+    PyObject *py_return_type, *py_param_types;
+    GType return_type;
+    guint n_params, i;
+    GType *param_types;
+    guint signal_id;
+
+    if (!PyArg_ParseTuple(tuple, "iOO", &signal_flags, &py_return_type,
+			  &py_param_types)) {
+	gchar buf[128];
+
+	PyErr_Clear();
+	g_snprintf(buf, sizeof(buf), "value for __gsignals__['%s'] not in correct format", signal_name);
+	PyErr_SetString(PyExc_TypeError, buf);
+	return FALSE;
+    }
+
+    return_type = pyg_type_from_object(py_return_type);
+    if (!return_type)
+	return FALSE;
+    if (!PySequence_Check(py_param_types)) {
+	gchar buf[128];
+
+	g_snprintf(buf, sizeof(buf), "third element of __gsignals__['%s'] tuple must be a sequence", signal_name);
+	PyErr_SetString(PyExc_TypeError, buf);
+	return FALSE;
+    }
+    n_params = PySequence_Length(py_param_types);
+    param_types = g_new(GType, n_params);
+    for (i = 0; i < n_params; i++) {
+	PyObject *item = PySequence_GetItem(py_param_types, i);
+
+	param_types[i] = pyg_type_from_object(item);
+	if (param_types[i] == 0) {
+	    Py_DECREF(item);
+	    g_free(param_types);
+	    return FALSE;
+	}
+	Py_DECREF(item);
+    }
+
+    signal_id = g_signal_newv(signal_name, instance_type, signal_flags,
+			      pyg_signal_class_closure_get(),
+			      (GSignalAccumulator)0, NULL,
+			      (GSignalCMarshaller)0,
+			      return_type, n_params, param_types);
+    g_free(param_types);
+
+    if (signal_id == 0) {
+	gchar buf[128];
+
+	g_snprintf(buf, sizeof(buf), "could not create signal for %s",
+		   signal_name);
+	PyErr_SetString(PyExc_RuntimeError, buf);
+	return FALSE;
+    }
+    return TRUE;
+}
+
+static gboolean
+override_signal(GType instance_type, const gchar *signal_name)
+{
+    guint signal_id;
+
+    signal_id = g_signal_lookup(signal_name, instance_type);
+    if (!signal_id) {
+	gchar buf[128];
+
+	g_snprintf(buf, sizeof(buf), "could not look up %s", signal_name);
+	PyErr_SetString(PyExc_TypeError, buf);
+	return FALSE;
+    }
+    g_signal_override_class_closure(signal_id, instance_type,
+				    pyg_signal_class_closure_get());
+    return TRUE;
+}
+
+static gboolean
+add_signals (GType instance_type, PyObject *signals)
+{
+    gboolean ret = TRUE;
+    GObjectClass *oclass;
+    int pos = 0;
+    PyObject *key, *value;
+
+    oclass = g_type_class_ref(instance_type);
+    while (PyDict_Next(signals, &pos, &key, &value)) {
+	const gchar *signal_name;
+
+	if (!PyString_Check(key)) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gsignals__ keys must be strings");
+	    ret = FALSE;
+	    break;
+	}
+	signal_name = PyString_AsString (key);
+
+	if (value == Py_None ||
+	    (PyString_Check(value) &&
+	     !strcmp(PyString_AsString(value), "override"))) {
+	    ret = override_signal(instance_type, signal_name);
+	} else {
+	    ret = create_signal(instance_type, signal_name, value);
+	}
+
+	if (!ret)
+	    break;
+    }
+    g_type_class_unref(oclass);
+    return ret;
+}
+
+static gboolean
+create_property (GObjectClass *oclass, const gchar *prop_name,
+		 GType prop_type, const gchar *nick, const gchar *blurb,
+		 PyObject *args, GParamFlags flags)
+{
+    GParamSpec *pspec = NULL;
+
+    switch (G_TYPE_FUNDAMENTAL(prop_type)) {
+    case G_TYPE_CHAR:
+	{
+	    gchar minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "ccc", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_char (prop_name, nick, blurb, minimum,
+				       maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_UCHAR:
+	{
+	    gchar minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "ccc", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_uchar (prop_name, nick, blurb, minimum,
+					maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_BOOLEAN:
+	{
+	    gboolean default_value;
+
+	    if (!PyArg_ParseTuple(args, "i", &default_value))
+		return FALSE;
+	    pspec = g_param_spec_boolean (prop_name, nick, blurb,
+					  default_value, flags);
+	}
+	break;
+    case G_TYPE_INT:
+	{
+	    gint minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "iii", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_int (prop_name, nick, blurb, minimum,
+				      maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_UINT:
+	{
+	    guint minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "iii", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_uint (prop_name, nick, blurb, minimum,
+				       maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_LONG:
+	{
+	    glong minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "lll", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_long (prop_name, nick, blurb, minimum,
+				       maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_ULONG:
+	{
+	    gulong minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "lll", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_ulong (prop_name, nick, blurb, minimum,
+					maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_INT64:
+	{
+	    gint64 minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "LLL", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_int64 (prop_name, nick, blurb, minimum,
+					maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_UINT64:
+	{
+	    guint64 minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "LLL", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_uint64 (prop_name, nick, blurb, minimum,
+					 maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_ENUM:
+	{
+	    gint default_value;
+
+	    if (!PyArg_ParseTuple(args, "i", &default_value))
+		return FALSE;
+	    pspec = g_param_spec_enum (prop_name, nick, blurb,
+				       prop_type, default_value, flags);
+	}
+	break;
+    case G_TYPE_FLAGS:
+	{
+	    guint default_value;
+
+	    if (!PyArg_ParseTuple(args, "i", &default_value))
+		return FALSE;
+	    pspec = g_param_spec_flags (prop_name, nick, blurb,
+					prop_type, default_value, flags);
+	}
+	break;
+    case G_TYPE_FLOAT:
+	{
+	    gfloat minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "fff", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_float (prop_name, nick, blurb, minimum,
+					maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_DOUBLE:
+	{
+	    gdouble minimum, maximum, default_value;
+
+	    if (!PyArg_ParseTuple(args, "ddd", &minimum, &maximum,
+				  &default_value))
+		return FALSE;
+	    pspec = g_param_spec_double (prop_name, nick, blurb, minimum,
+					 maximum, default_value, flags);
+	}
+	break;
+    case G_TYPE_STRING:
+	{
+	    const gchar *default_value;
+
+	    if (!PyArg_ParseTuple(args, "z", &default_value))
+		return FALSE;
+	    pspec = g_param_spec_string (prop_name, nick, blurb,
+					 default_value, flags);
+	}
+	break;
+    case G_TYPE_PARAM:
+	if (!PyArg_ParseTuple(args, ""))
+	    return FALSE;
+	pspec = g_param_spec_param (prop_name, nick, blurb, prop_type, flags);
+	break;
+    case G_TYPE_BOXED:
+	if (!PyArg_ParseTuple(args, ""))
+	    return FALSE;
+	pspec = g_param_spec_boxed (prop_name, nick, blurb, prop_type, flags);
+	break;
+    case G_TYPE_POINTER:
+	if (!PyArg_ParseTuple(args, ""))
+	    return FALSE;
+	pspec = g_param_spec_pointer (prop_name, nick, blurb, flags);
+	break;
+    case G_TYPE_OBJECT:
+	if (!PyArg_ParseTuple(args, ""))
+	    return FALSE;
+	pspec = g_param_spec_object (prop_name, nick, blurb, prop_type, flags);
+	break;
+    default:
+	/* unhandled pspec type ... */
+	break;
+    }
+
+    if (!pspec) {
+	char buf[128];
+
+	g_snprintf(buf, sizeof(buf), "could not create param spec for type %s",
+		   g_type_name(prop_type));
+	PyErr_SetString(PyExc_TypeError, buf);
+	return FALSE;
+    }
+    g_object_class_install_property(oclass, 1, pspec);
+    return TRUE;
+}
+
+static gboolean
+add_properties (GType instance_type, PyObject *properties)
+{
+    gboolean ret = TRUE;
+    GObjectClass *oclass;
+    int pos = 0;
+    PyObject *key, *value;
+
+    oclass = g_type_class_ref(instance_type);
+    while (PyDict_Next(properties, &pos, &key, &value)) {
+	const gchar *prop_name;
+	GType prop_type;
+	const gchar *nick, *blurb;
+	GParamFlags flags;
+	gint val_length;
+	PyObject *slice, *item, *py_prop_type;
+	/* values are of format (type,nick,blurb, type_specific_args, flags) */
+	
+	if (!PyString_Check(key)) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gproperties__ keys must be strings");
+	    ret = FALSE;
+	    break;
+	}
+	prop_name = PyString_AsString (key);
+
+	if (!PyTuple_Check(value)) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gproperties__ values must be tuples");
+	    ret = FALSE;
+	    break;
+	}
+	val_length = PyTuple_Size(value);
+	if (val_length < 4) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gproperties__ values must be at least 4 elements long");
+	    ret = FALSE;
+	    break;
+	}	    
+
+	slice = PySequence_GetSlice(value, 0, 3);
+	if (!slice) {
+	    ret = FALSE;
+	    break;
+	}
+	if (!PyArg_ParseTuple(slice, "Ozz", &py_prop_type, &nick, &blurb)) {
+	    Py_DECREF(slice);
+	    ret = FALSE;
+	    break;
+	}
+	Py_DECREF(slice);
+	prop_type = pyg_type_from_object(py_prop_type);
+	if (!prop_type) {
+	    ret = FALSE;
+	    break;
+	}
+	item = PyTuple_GetItem(value, val_length-1);
+	if (!PyInt_Check(item)) {
+	    PyErr_SetString(PyExc_TypeError,
+		"last element in __gproperties__ value tuple must be an int");
+	    ret = FALSE;
+	    break;
+	}
+	flags = PyInt_AsLong(item);
+
+	/* slice is the extra items in the tuple */
+	slice = PySequence_GetSlice(value, 3, val_length-1);
+	ret = create_property(oclass, prop_name, prop_type, nick, blurb,
+			      slice, flags);
+	Py_DECREF(slice);
+
+	if (!ret)
+	    break;
+    }
+    g_type_class_unref(oclass);
+    return ret;
+}
+
 static PyObject *
 pyg_type_register(PyObject *self, PyObject *args)
 {
-    PyObject *gtype, *module;
+    PyObject *gtype, *module, *gsignals, *gproperties;
     PyTypeObject *class;
     GType parent_type, instance_type;
     gchar *type_name = NULL;
@@ -1940,7 +2512,7 @@ pyg_type_register(PyObject *self, PyObject *args)
 	(GBaseInitFunc) NULL,
 	(GBaseFinalizeFunc) NULL,
 
-	(GClassInitFunc) NULL,
+	(GClassInitFunc) pyg_object_class_init,
 	(GClassFinalizeFunc) NULL,
 	NULL, /* class_data */
 
@@ -1979,6 +2551,9 @@ pyg_type_register(PyObject *self, PyObject *args)
 	if (type_name[i] == '.')
 	    type_name[i] = '+';
 
+    /* set class_data that will be passed to the class_init function. */
+    type_info.class_data = class;
+
     /* fill in missing values of GTypeInfo struct */
     g_type_query(parent_type, &query);
     type_info.class_size = query.class_size;
@@ -1993,11 +2568,55 @@ pyg_type_register(PyObject *self, PyObject *args)
 	return NULL;
     }
 
+    /* store pointer to the class with the GType */
+    Py_INCREF(class);
+    g_type_set_qdata(instance_type, pygobject_class_key, class);
+
     /* set new value of __gtype__ on class */
     gtype = pyg_type_wrapper_new(instance_type);
     PyDict_SetItemString(class->tp_dict, "__gtype__", gtype);
     Py_DECREF(gtype);
 
+    /* we look this up in the instance dictionary, so we don't
+     * accidentally get a parent type's __gsignals__ attribute. */
+    gsignals = PyDict_GetItemString(class->tp_dict, "__gsignals__");
+    if (gsignals) {
+	if (!PyDict_Check(gsignals)) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gsignals__ attribute not a dict!");
+	    Py_DECREF(gsignals);
+	    return NULL;
+	}
+	if (!add_signals(instance_type, gsignals)) {
+	    Py_DECREF(gsignals);
+	    return NULL;
+	}
+	PyDict_DelItemString(class->tp_dict, "__gsignals__");
+	Py_DECREF(gsignals);
+    } else {
+	PyErr_Clear();
+    }
+
+    /* we look this up in the instance dictionary, so we don't
+     * accidentally get a parent type's __gsignals__ attribute. */
+    gproperties = PyDict_GetItemString(class->tp_dict, "__gproperties__");
+    if (gproperties) {
+	if (!PyDict_Check(gproperties)) {
+	    PyErr_SetString(PyExc_TypeError,
+			    "__gproperties__ attribute not a dict!");
+	    Py_DECREF(gproperties);
+	    return NULL;
+	}
+	if (!add_properties(instance_type, gproperties)) {
+	    Py_DECREF(gproperties);
+	    return NULL;
+	}
+	PyDict_DelItemString(class->tp_dict, "__gproperties__");
+	Py_DECREF(gproperties);
+    } else {
+	PyErr_Clear();
+    }
+
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -2310,31 +2929,36 @@ pyg_flags_add_constants(PyObject *module, GType flags_type,
     g_type_class_unref(fclass);
 }
 
-static int
-pyg_fatal_exceptions_notify(void)
+static gboolean
+pyg_error_check(GError **error)
 {
-    GList *tmp_list = pygobject_exception_notifiers;
-    if (!tmp_list)
-	return 0;
-    while (tmp_list != NULL) {
-	PyGFatalExceptionFunc notifier = tmp_list->data;
-	(* notifier)();
-	tmp_list = g_list_next (tmp_list);
-    }
-    return 1;
-}
-static void
-pyg_fatal_exceptions_notify_add(PyGFatalExceptionFunc func)
-{
-    pygobject_exception_notifiers = 
-	g_list_append(pygobject_exception_notifiers, func);
-}
+    g_return_val_if_fail(error != NULL, FALSE);
 
-static void
-pyg_fatal_exceptions_notify_remove(PyGFatalExceptionFunc func)
-{
-    pygobject_exception_notifiers = 
-	g_list_remove(pygobject_exception_notifiers, func);
+    if (*error != NULL) {
+	PyObject *exc_instance;
+	PyObject *d;
+
+	exc_instance = PyObject_CallFunction(gerror_exc, "z",
+					     (*error)->message);
+	PyObject_SetAttrString(exc_instance, "domain",
+			       d=PyString_FromString(g_quark_to_string((*error)->domain)));
+	Py_DECREF(d);
+	PyObject_SetAttrString(exc_instance, "code",
+			       d=PyInt_FromLong((*error)->code));
+	Py_DECREF(d);
+	if ((*error)->message) {
+	    PyObject_SetAttrString(exc_instance, "message",
+				   d=PyString_FromString((*error)->message));
+	    Py_DECREF(d);
+	} else {
+	    PyObject_SetAttrString(exc_instance, "message", Py_None);
+	}
+
+	PyErr_SetObject(gerror_exc, exc_instance);
+	g_clear_error(error);
+	return TRUE;
+    }
+    return FALSE;
 }
 
 /* ----------------- gobject module initialisation -------------- */
@@ -2362,16 +2986,15 @@ static struct _PyGObject_Functions functions = {
   pyg_enum_add_constants,
   pyg_flags_add_constants,
   
-  pyg_fatal_exceptions_notify_add,
-  pyg_fatal_exceptions_notify_remove,
-
   pyg_constant_strip_prefix,
+
+  pyg_error_check,
 };
 
 DL_EXPORT(void)
 initgobject(void)
 {
-    PyObject *m, *d, *o;
+    PyObject *m, *d, *o, *tuple;
 
     PyGTypeWrapper_Type.ob_type = &PyType_Type;
     PyGParamSpec_Type.ob_type = &PyType_Type;
@@ -2381,10 +3004,17 @@ initgobject(void)
 
     g_type_init();
 
+    pygobject_class_key = g_quark_from_static_string("PyGObject::class");
+    pygobject_wrapper_key = g_quark_from_static_string("PyGObject::wrapper");
+    pygobject_ownedref_key = g_quark_from_static_string("PyGObject::ownedref");
+
     PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
 						  pyobject_copy,
 						  pyobject_free);
 
+    gerror_exc = PyErr_NewException("gobject.GError", PyExc_RuntimeError,NULL);
+    PyDict_SetItemString(d, "GError", gerror_exc);
+
     pygobject_register_class(d, "GObject", G_TYPE_OBJECT,
 			     &PyGObject_Type, NULL);
 
@@ -2406,8 +3036,10 @@ initgobject(void)
 
     boxed_marshalers = g_hash_table_new(g_direct_hash, g_direct_equal);
 
-    pygobject_wrapper_key = g_quark_from_static_string("py-gobject-wrapper");
-    pygobject_ownedref_key = g_quark_from_static_string("py-gobject-ownedref");
+    tuple = Py_BuildValue ("(iii)", glib_major_version, glib_minor_version,
+			   glib_micro_version);
+    PyDict_SetItemString(d, "glib_version", tuple);    
+    Py_DECREF(tuple);
 
     /* for addon libraries ... */
     PyDict_SetItemString(d, "_PyGObject_API",
@@ -2422,6 +3054,13 @@ initgobject(void)
     PyModule_AddIntConstant(m, "SIGNAL_ACTION", G_SIGNAL_ACTION);
     PyModule_AddIntConstant(m, "SIGNAL_NO_HOOKS", G_SIGNAL_NO_HOOKS);
 
+    PyModule_AddIntConstant(m, "PARAM_READABLE", G_PARAM_READABLE);
+    PyModule_AddIntConstant(m, "PARAM_WRITABLE", G_PARAM_WRITABLE);
+    PyModule_AddIntConstant(m, "PARAM_CONSTRUCT", G_PARAM_CONSTRUCT);
+    PyModule_AddIntConstant(m, "PARAM_CONSTRUCT_ONLY", G_PARAM_CONSTRUCT_ONLY);
+    PyModule_AddIntConstant(m, "PARAM_LAX_VALIDATION", G_PARAM_LAX_VALIDATION);
+    PyModule_AddIntConstant(m, "PARAM_READWRITE", G_PARAM_READWRITE);
+
     PyModule_AddObject(m, "TYPE_INVALID", pyg_type_wrapper_new(G_TYPE_INVALID));
     PyModule_AddObject(m, "TYPE_NONE", pyg_type_wrapper_new(G_TYPE_NONE));
     PyModule_AddObject(m, "TYPE_INTERFACE", pyg_type_wrapper_new(G_TYPE_INTERFACE));
@@ -2432,6 +3071,8 @@ initgobject(void)
     PyModule_AddObject(m, "TYPE_UINT", pyg_type_wrapper_new(G_TYPE_UINT));
     PyModule_AddObject(m, "TYPE_LONG", pyg_type_wrapper_new(G_TYPE_LONG));
     PyModule_AddObject(m, "TYPE_ULONG", pyg_type_wrapper_new(G_TYPE_ULONG));
+    PyModule_AddObject(m, "TYPE_INT64", pyg_type_wrapper_new(G_TYPE_INT64));
+    PyModule_AddObject(m, "TYPE_UINT64", pyg_type_wrapper_new(G_TYPE_UINT64));
     PyModule_AddObject(m, "TYPE_ENUM", pyg_type_wrapper_new(G_TYPE_ENUM));
     PyModule_AddObject(m, "TYPE_FLAGS", pyg_type_wrapper_new(G_TYPE_FLAGS));
     PyModule_AddObject(m, "TYPE_FLOAT", pyg_type_wrapper_new(G_TYPE_FLOAT));
